import numpy as np
from skimage import io
from skimage.transform import resize
from skimage.feature import canny
from skimage.filters import sobel
from skimage.transform import hough_circle, hough_circle_peaks, hough_line, hough_line_peaks
from skimage.util import img_as_ubyte
from glob import glob
from .diagnose import getImageStats
import re
import math

def hasExactlyOneLine(img: np.ndarray, sigma: int = 0, low_threshold:float = None, high_threshold: float=None):
    """Detects the presence of a vertical or horizontal edge in an image. 
    To be used when filtering out edge-tiles from a dataset.

    Parameters
    ----------
    img : np.ndarray
        img
    sigma : int
        sigma to be used for gaussian blurring by canny edge filter. Default = 0
        For more documentation, see https://scikit-image.org/docs/dev/api/skimage.feature.html
    low_threshold : float
        Low threshold to be used by hysteresis thresholding. 
        For more documentation, see https://scikit-image.org/docs/dev/api/skimage.feature.html
    high_threshold : float
        High threshold to be used by hysteresis thresholding. 
        For more documentation, see https://scikit-image.org/docs/dev/api/skimage.feature.html
    """

    edge_image = canny(img, sigma=0, low_threshold = low_threshold, high_threshold = high_threshold)
    tested_angles = np.array([np.pi, np.pi / 2])
    h, theta, d = hough_line(edge_image, theta=tested_angles)
    accums, angles, dists =  hough_line_peaks(h, theta, d)
    return len(accums) == 1

def hasExactlyOneCircle(img: np.ndarray, min_ydistance:int = 50, min_xdistance:int = 50):
    """Checks whether the input image contains exactly one circle-like object
    Note: For this, it is important that the input img is 16bit.

    Parameters
    ----------
    img : np.ndarray
        img
    min_ydistance : int
        Min ydistance that two circles to be detected can be apart.
    min_xdistance : int
        Min xdistance that two circles to be detected can be apart.
    """

    sobel_image = sobel(img)
    sobel_image = img_as_ubyte(sobel_image)
    hough_radii = np.arange(25, 75, 2)
    hough_res = hough_circle(sobel_image, hough_radii)
    accums, cx, cy, radii = hough_circle_peaks(hough_res, hough_radii, min_ydistance = 50, min_xdistance= 50 )
    return len(radii) == 1


def isGoodQuality(image: np.ndarray):
    df = getImageStats(image, out_print=False)
    if df["Actual range"][0] == (0,0):
        return False
    elif df["Actual range"][0][0] != 0:
        return False
    elif df["Otsu threshold"][0] < 0:
        return False
    elif float(df["Laplace std"][0]) < 0.001:
        return False
    elif df["Otsu threshold"][0] < 100:
        return False 
    elif df["Otsu threshold"][0] > 1500:
        return False 
    else:
        return True
if __name__ == "__main__":
    from sklearn.metrics import confusion_matrix
    import seaborn as sns
    import matplotlib.pyplot as plt     
    full_image_list = glob("/home/david/.config/nnn/mounts/nacho@10.38.76.144/tool/enteric_neurones/slidescanner_examples/tiles/*/*.tif")
    y_labs = ["good" in file for file in full_image_list ]
    y_preds =  [isGoodQuality(io.imread(file)) for file in full_image_list]
    correct_array = [pred == label for pred, label in zip(y_preds, y_labs)]
    wrong_images = [b for a, b in zip(correct_array, full_image_list) if not a]
    right_images = [b for a, b in zip(correct_array, full_image_list) if a]
    right_bad_images = [b for a, b in zip(correct_array, full_image_list) if a and "bad" in b]
    right_good_images = [b for a, b in zip(correct_array, full_image_list) if a and "good" in b]
    wrong_bad_images = [b for a, b in zip(correct_array, full_image_list) if not a and "bad" in b]
    wrong_good_images = [b for a, b in zip(correct_array, full_image_list) if not a and "good" in b]

    to_plot = wrong_good_images
    title = "wrong_good_images"
    # Testing dynamic plotting
    tot = len(to_plot)
    cols = round(math.sqrt(tot))
    rows = tot // cols 
    rows += tot % cols

    position = range(1, tot+1)

    fig = plt.figure(1)
    st = fig.suptitle(title, fontsize="x-large")
    for k in range(tot):
        ax = fig.add_subplot(rows, cols, position[k])
        image = io.imread(to_plot[k])
        ax.imshow(resize(image, (image.shape[0] // 5, image.shape[1] // 5),anti_aliasing=True))
        ax.set_title(re.findall(r'Slide\d+-\d+-\d+', to_plot[k])[0]  + "-"  + re.findall(r'tile\d+', to_plot[k])[0])
        ax.axis("off")
    st.set_y(0.95)
    fig.subplots_adjust(top=0.85)
    plt.show()

    # Confusion matrix
    # cm = confusion_matrix(y_labs, y_preds, labels=[True, False])
    # ax= plt.subplot()
    # sns.heatmap(cm, annot=True, fmt='g', ax=ax);  #annot=True to annotate cells, ftm='g' to disable scientific notation

    # # labels, title and ticks
    # ax.set_xlabel('Predicted labels');ax.set_ylabel('True labels'); 
    # ax.set_title('Confusion Matrix'); 
    # ax.xaxis.set_ticklabels(['Good', 'Bad']); ax.yaxis.set_ticklabels(['Good', 'Bad']);

    # plt.show()



import os
from skimage import io
from skimage.filters import laplace
import numpy as np
from typing import List, Tuple
import aicspylibczi
from nd2reader import ND2Reader
from itertools import product
from .utils import getLastDimensions

def getMostInFocusImage(image_array_list: List[np.array]) -> Tuple[np.array, int]:  
    """Gets most in focus image and its index from a list of image-arrays.

    Parameters
    ----------
    image_array_list : List[np.array]
        image_array_list

    Returns
    -------
    Tuple[np.array, int]

    """
    stdev_list = []
    for image in image_array_list:
        # calculate edges in image
        edged = laplace(image)
        # Calculate stdev of edges
        stdev = np.std(edged)
        stdev_list.append(stdev)
    # Find largest stdev in list
    largest = max(stdev_list)
    # Fidn which index it is to link back to the original list
    index = stdev_list.index(largest)
    return image_array_list[index], index

def readImage(file_path: str) -> Tuple[np.array, str]:
    """Reads in image-file and returns both the image, its file type and its shape.

    Parameters
    ----------
    file_path : str
        path to image

    Returns
    -------
    Tuple[np.array, str]

    """
    if file_path.lower().endswith((".tif", ".tiff")):
        image = io.imread(file_path)
        img_type = "tif"
        img_shape = image.shape
        return image, img_type, img_shape
    if file_path.lower().endswith(".czi"):
        image = aicspylibczi.CziFile(file_path)
        img_type = "czi"
        img_shape  = image.get_dims_shape()[0]
        return image, img_type, img_shape
    if file_path.lower().endswith(".nd2"):
        with ND2Reader(file_path) as image:
            img_type = "nd2"
            img_shape = image.sizes
            return file_path, img_type, img_shape


def extractMostInFocusZstack(image: np.ndarray, z_shape_index:int = -1) -> np.ndarray:
    """Extract the most in focus z-stack of an image.

    Parameters
    ----------
    image : np.ndarray
        image
    z_shape_index : int
        index that indicates which position in the image.shape contains the z-axis.

    Returns
    -------
    np.ndarray

    """
    if not len(image.shape) > 2:
        raise ValueError(f"Image shape expected to be larger than 2, image shape is {image.shape}")
    img_list = []
    for i in range(image.shape[z_shape_index]):
        img_list.append(np.take(image, i, z_shape_index))
    mostInFocusImage, index = getMostInFocusImage(img_list)
    return mostInFocusImage, index



def extractSingleImages(image, image_type: str, indexes_dict: dict = None, filename: str = ""):
    """Extract from multi-dimensional input images all singe-stack images indicated by the indexes dict.

    Parameters
    ----------
    image : np.ndarray
        multi-dimensional input image
    indexes_dict : dict
        dict where keys are the shape-indexes and their values are indexes of which channels in that axis to extract. The combination of all key-value pairs will be extracted into single-channel images. Default = all combinations of axises that aren't the first 2 (which usually represents X and Y).
    image_type : str
        image_type, choose either "tif", "czi" or "nd2"
    filename : str
        filename
    """
    if image_type == "tif":
        # First check if it's a string, if so we gotta read the image still
        if isinstance(image, str):
            if not filename: # If no filename is given, we'll take the input image string
                filename = image
            image = io.imread(image)
        if indexes_dict == None: # if no indexes_dict is given, calculate all combinations of no X and Y axises
            indexes_list = [list(range(image.shape[i])) for i in range(len(image.shape))[2:]]
            # Create combinations of all channels
            all_combinations =list(product(*indexes_list))
            for combo in all_combinations:
                tmp_extracted_image = np.copy(image) # create copy to iteratively extract dimensions
                append_string = "" # String to append to the end of this extracted img
                for i, entry in enumerate(combo,2):
                    tmp_extracted_image = np.take(tmp_extracted_image, entry, 2)
                    append_string += f"i{i}_c{entry}"
                io.imsave(f"{os.path.splitext(filename)[0]}_{append_string}.tif", tmp_extracted_image, check_contrast=False)
        else:
            for shape_index, channel_indexes in indexes_dict.items():
                if isinstance(channel_indexes, int): # if it's only one channel, it'ss give an iterable error, so we gotta make it a list
                    channel_indexes = [channel_indexes] 
                for channel_index in channel_indexes:
                    if shape_index == None:
                        extracted_image = np.take(image, channel_index, axis=-1)
                    else:
                        extracted_image = np.take(image, channel_index, axis=shape_index)
                    io.imsave(f"{os.path.splitext(filename)[0]}_i{shape_index}_c{channel_index}.tif", extracted_image)

            
    elif image_type == "czi":
        if isinstance(image, str):
            if not filename: # If no filename is given, we'll take the input image string
                filename = image
            image = aicspylibczi.CziFile(image)
        if indexes_dict == None: # if no indexes_dict is given, calculate all combinations of no X and Y axises
            channel_indices = range(*image.get_dims_shape()[0]["C"])
            for channel_index in channel_indices:
                if image.is_mosaic():

                    tmp_image = image.read_mosaic(C=channel_index,scale_factor=1)
                else:
                    tmp_image = image.read_image(C=channel_index)

                tmp_image = getLastDimensions(tmp_image, nr_dims = 2)# extract last 2 dims, first or second dim might be T or S
                io.imsave(f"{os.path.splitext(filename)[0]}_c{channel_index}.tif", tmp_image, check_contrast=False)

        else:
            for shape_index, channel_indexes in indexes_dict.items():
                # now we assume the shape indexes are letters referring to czi stuff
                if isinstance(channel_indexes, int): # if it's only one channel, it'ss give an iterable error, so we gotta make it a list
                    channel_indexes = [channel_indexes] 
                for channel_index in channel_indexes:
                    extracted_image,shape = image.read_image(shape_index=channel_index)
                    extracted_image = getLastDimensions(extracted_image, nr_dims = 2)
                    io.imsave(f"{os.path.splitext(filename)[0]}__{shape_index}{channel_index}.tif", extracted_image )

    elif image_type == "nd2":
        with ND2Reader(image) as images:
            for shape_index, channel_indexes in indexes_dict.items():
                if isinstance(channel_indexes, int): # if it's only one channel, it'ss give an iterable error, so we gotta make it a list
                    channel_indexes = [channel_indexes] 
                for channel_index in channel_indexes:
                    extracted_image = images.get_frame_2D(c=channel_index)
                    io.imsave(f"{os.path.splitext(filename)[0]}_i{shape_index}_c{channel_index}.tif", extracted_image)



import numpy as np

def otsuThreshold(image: np.array) -> float:
    """Use otsu's thresholding to find the optimal threshold divind the histogram of the image into 2 distributions.

    Parameters
    ----------
    image : np.array
        image

    Returns
    -------
    float

    """
    bins_num = 256

    # Get the image histogram
    hist, bin_edges = np.histogram(image, bins=bins_num)

    hist = hist.astype(float)

    # # Get normalized histogram if it is required
    # hist = np.divide(hist.ravel(), hist.max())

    # # Calculate centers of bins
    bin_mids = (bin_edges[:-1] + bin_edges[1:]) / 2.

    # Iterate over all thresholds (indices) and get the probabilities w1(t), w2(t)
    weight1 = np.cumsum(hist)
    weight2 = np.cumsum(hist[::-1])[::-1]

    # Get the class means mu0(t)
    mean1 = np.cumsum(hist * bin_mids) / weight1
    # Get the class means mu1(t)
    mean2 = (np.cumsum((hist * bin_mids)[::-1]) / weight2[::-1])[::-1]

    inter_class_variance = weight1[:-1] * weight2[1:] * (mean1[:-1] - mean2[1:]) ** 2

    # Maximize the inter_class_variance function val
    index_of_max_val = np.argmax(inter_class_variance)

    threshold = bin_mids[:-1][index_of_max_val]
    return threshold



import time
from joblib import Parallel, delayed
import concurrent.futures
import os
from functools import wraps

def measureTime(func):
    def wrapper(*args, **kwargs):
        starttime = time.perf_counter()
        temp = func(*args, **kwargs)
        endtime = time.perf_counter()
        print(f"Time needed to run {func.__name__}: {time.strftime('%H:%M:%S', time.gmtime(endtime - starttime))}")
        return(temp)
    return wrapper


# source: https://medium.com/analytics-vidhya/python-decorator-to-parallelize-any-function-23e5036fb6a
def makeParallel(func):
    """
        Decorator used to decorate any function which needs to be parallized.
        After the input of the function should be a list in which each element is a instance of input fol the normal function.
        You can also pass in keyword arguments seperateley.
        :param func: function
            The instance of the function that needs to be parallelized.
        :return: function


        Usage:
        iterable = list(range(1, 500))
        results = makeParallel(sample_function)(list_of_post_ids, args)

    """

    @wraps(func)
    def wrapper(lst):
        """
        :param lst:
            The inputs of the function in a list.
        :return:
        """
        # the number of threads that can be max-spawned.
        # If the number of threads are too high, then the overhead of creating the threads will be significant.
        # Here we are choosing the number of CPUs available in the system and then multiplying it with a constant.
        # In my system, i have a total of 8 CPUs so i will be generating a maximum of 16 threads in my system.
        number_of_threads_multiple = 2 # You can change this multiple according to you requirement
        number_of_workers = int(os.cpu_count() * number_of_threads_multiple)
        if len(lst) < number_of_workers:
            # If the length of the list is low, we would only require those many number of threads.
            # Here we are avoiding creating unnecessary threads
            number_of_workers = len(lst)

        if number_of_workers:
            if number_of_workers == 1:
                # If the length of the list that needs to be parallelized is 1, there is no point in
                # parallelizing the function.
                # So we run it serially.
                result = [func(lst[0])]
            else:
                # Core Code, where we are creating max number of threads and running the decorated function in parallel.
                result = []
                with concurrent.futures.ThreadPoolExecutor(max_workers=number_of_workers) as executer:
                    bag = {executer.submit(func, i): i for i in lst}
                    for future in concurrent.futures.as_completed(bag):
                        result.append(future.result())
        else:
            result = []
        return result
    return wrapper



from __future__ import print_function, unicode_literals, absolute_import, division
import numpy as np
import cv2
from skimage import io
from skimage.util import img_as_ubyte
from skimage import color
import matplotlib.pyplot as plt
from skimage import measure
from csbdeep.utils import Path, normalize
from stardist.models import StarDist2D
from .display import showSegmentation
from cellpose import models


def otsuSegment(img: np.array) -> np.array :
    '''
        note to self: this code adapted from DigitalSreeni assumes that your input image is an 8-bit rgb, which makes it so that we have to do some image format transformation because:
        - cv2.shreshold accepts only 8-bit grayscale
        - cv2.watershed only accepts 8-bit rgb
    '''
    '''
    returns a labeled image, where 0 = background and all other integers an object number.
    These numbers don't have any actual image value, so the image isn't really used as an image object, but more as an array
    that stores which pixel belongs to which label. Also returns a csv that contains image properties of the given objects
    '''
    # Create an 8bit version of the image
    img_as_8 = img_as_ubyte(img)
    # Creat an RGB version that only has signal in the blue channel
    shape = img_as_8.shape
    empty = np.zeros(shape)
    img_as_8bit_RGB = cv2.merge([img_as_8,img_as_8,img_as_8])
    try:
        cells = img[:,:,0]
    except IndexError:
        cells = img_as_8

    ret1, thresh = cv2.threshold(cells, 0, 255, cv2.THRESH_BINARY+cv2.THRESH_OTSU)

    # Morphological operations to remove small noise - opening
    #To remove holes we can use closing
    kernel = np.ones((3,3),np.uint16)
    opening = cv2.morphologyEx(thresh,cv2.MORPH_OPEN,kernel, iterations = 2)

    # from skimage.segmentation import clear_border
    # opening = clear_border(opening) #Remove edge touching grains

    sure_bg = cv2.dilate(opening,kernel,iterations=10)
    dist_transform = cv2.distanceTransform(opening,cv2.DIST_L2,5)
    ret2, sure_fg = cv2.threshold(dist_transform,0.5*dist_transform.max(),255,0)

    #Later you realize that 0.25* max value will not separate the cells well.
    #High value like 0.7 will not recognize some cells. 0.5 seems to be a good compromize

    # Unknown ambiguous region is nothing but bkground - foreground
    sure_fg = np.uint8(sure_fg)
    unknown = cv2.subtract(sure_bg,sure_fg)

    #Now we create a marker and label the regions inside.
    # For sure regions, both foreground and background will be labeled with positive numbers.
    # Unknown regions will be labeled 0.
    #For markers let us use ConnectedComponents.
    ret3, markers = cv2.connectedComponents(sure_fg)

    #One problem rightnow is that the entire background pixels is given value 0.
    #This means watershed considers this region as unknown.
    #So let us add 10 to all labels so that sure background is not 0, but 10
    markers = markers+10

    # Now, mark the region of unknown with zero
    markers[unknown==255] = 0

    #Now we are ready for watershed filling.
    markers = cv2.watershed(img_as_8bit_RGB,markers)

    #The boundary region will be marked -1
    markers[markers==-1] = 10 # add the boundary images to the background

    label_image = measure.label(markers, background=10)
    return label_image


def stardistSegment(img: np.array, model:str = "2D_versatile_fluo") -> np.array:
    """Segments cells using pretrained stardist models.

    Parameters
    ----------
    img : np.array
        image to be segmented
    model : str
        String representing which pretrained stardist model to load. Choose one of: (2D_versatile_fluo, 2D_versatile_he, 2D_paper_dsb2018)

    Returns
    -------
    np.array

    """
    model_versatile = StarDist2D.from_pretrained(model)

    # extract number of channels in case the input image is an RGB image
    n_channel = 1 if img.ndim == 2 else img.shape[-1]
    # depending on that, we want to normalize the channels independantly
    axis_norm = (0,1)   # normalize channels independently
    # axis_norm = (0,1,2) # normalize channels jointly

    img_normalized = normalize(img, 1,99.8, axis=axis_norm)

    labeled_image, details = model_versatile.predict_instances(img_normalized)
    return labeled_image 

def cellPoseSegment(img: np.array, model:str = "nuclei", channels: np.array = [0,0]) -> np.array:
    """Segments cells using pretrained Cellpose models. Can do both cytoplasm and nuclei at the same time.

    Parameters
    ----------
    img : np.array
        img
    model : str
        String representing which model to use. Either 'nuclei', 'cyto'
    channels : np.array
        Array depicting which channel is either nucleus or DAPI, default = [0,0] -> grayscale for all images
        How to indicate:

        Grayscale=0, R=1, G=2, B=3
        channels = [cytoplasm, nucleus]
        if NUCLEUS channel does not exist, set the second channel to 0
        IF ALL YOUR IMAGES ARE THE SAME TYPE, you can give a list with 2 elements
        channels = [0,0] # IF YOU HAVE GRAYSCALE
        channels = [2,3] # IF YOU HAVE G=cytoplasm and B=nucleus
        channels = [2,1] # IF YOU HAVE G=cytoplasm and R=nucleus
        If they have different orders -> make it a 2D array, e.g: 3 images:
        channels = [[2,3], [1,2], [3,1]]

    Returns
    -------
    np.array

    """
    # model_type='cyto' or model_type='nuclei'
    model = models.Cellpose(gpu=False, model_type=model)
    masks, flows, styles, diams = model.eval(img, diameter=None, channels=channels)
    return masks




import cv2
import numpy as np
from skimage.transform import resize
from skimage.morphology import white_tophat, disk
from skimage.util import img_as_uint
from skimage import io


# testing imports
import matplotlib.pyplot as plt
from skimage.draw import ellipse


# Automatic brightness and contrast optimization with optional histogram clipping
def automaticBrightnessAndContrast(image: np.array, clip_hist_percent: int =1):
    if np.amin(image) == 0 and np.amax(image) == 0:
        return image, 0, 0
    if len(image.shape) > 2:
        gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    else: 
        gray = image
    
    # Calculate grayscale histogram
    hist = cv2.calcHist([gray],[0],None,[256],[0,256])
    hist_size = len(hist)
    
    # Calculate cumulative distribution from the histogram
    accumulator = []
    accumulator.append(float(hist[0]))
    for index in range(1, hist_size):
        accumulator.append(accumulator[index -1] + float(hist[index]))
    
    # Locate points to clip
    maximum = accumulator[-1]
    clip_hist_percent *= (maximum/100.0)
    clip_hist_percent /= 2.0
    
    try:
        # Locate left cut
        minimum_gray = 0
        while accumulator[minimum_gray] < clip_hist_percent:
            minimum_gray += 1
        
        # Locate right cut
        maximum_gray = hist_size -1
        while accumulator[maximum_gray] >= (maximum - clip_hist_percent):
            maximum_gray -= 1
    except:
        return image
    
    # If the image was empty, the alpha calculation will raise an error, so just return the original image
    if maximum_gray - minimum_gray == 0:
        return image
    # Calculate alpha and beta values
    alpha = 255 / (maximum_gray - minimum_gray)
    beta = -minimum_gray * alpha

    
    '''
    # Calculate new histogram with desired range and show histogram 
    new_hist = cv2.calcHist([gray],[0],None,[256],[minimum_gray,maximum_gray])
    plt.plot(hist)
    plt.plot(new_hist)
    plt.xlim([0,256])
    plt.show()
    '''

    auto_result = cv2.convertScaleAbs(image, alpha=alpha, beta=beta)
    return auto_result

def equalizeImageSize(ref_image: np.array, target_image: np.array, save = False, out_name = "") -> np.array:
    # Gross ifelse statement to check if one of the image is globally larger or smaller than the other
    ref_shape, target_shape = ref_image.shape, target_image.shape
    # if (ref_shape[0] > target_shape[0] and ref_shape[1] < target_shape[1]) or (target_shape[0] > ref_shape[0] and target_shape[1] < ref_shape[1]):
    #     return ref_image, target_image
    # else:
    target_resized = resize(target_image,ref_image.shape)
    target_resized = img_as_uint(target_resized)
    if save and out_name:
        io.imsave(f"{out_name}_resized.tif", target_resized)
    return ref_image, target_resized

def whiteTophat(image, radius):
    """Wrapper for skimage's white tophat filter

    Parameters
    ----------
    image : np.ndarray
        Image to be filtered.
    radius: int
        Radius of the morphological disk.
    """
    selem = disk(radius)
    return white_tophat(image,selem)

def changeRGBcolor(image, original_color, target_color):
    """ Change a specific color in an image to another color

    Parameters
    ----------
    image :
        input image
    original_color :
        tuple representing the rgb color to change
    target_color :
        tuple representing the rgb color to change to 
    """
    image = image.copy()
    if len(image.shape) > 3:
        raise ValueError("not an RGB image")
    if image.shape[2] > 3:
        image = image[:,:,0:3]
    # print(image[:,:,0])
    r1, g1, b1 = original_color
    r2, g2, b2 = target_color

    red, green, blue = image[:,:,0], image[:,:,1], image[:,:,2]
    mask = (red == r1) & (green == g1) & (blue == b1)
    image[:,:,:3][mask] = [r2, g2, b2]
    return image

    # io.imsave("./image_fixed.png", data, check_contrast=False)

if __name__ == '__main__':
    ref_image = np.zeros((1600,500), dtype=np.uint)
    ref_image[ellipse(800,250, 700, 100)] = 255
    target_image = np.zeros((1000,1000), dtype=np.uint)
    target_image[ellipse(500,500, 400,200)] = 255

    warped_ref_image, warped_target_image = equalizeImageSize(ref_image, target_image)

    fig, axs = plt.subplots(2,2)
    axs[0,0].imshow(ref_image)
    axs[0,1].imshow(target_image)
    axs[1,0].imshow(warped_ref_image)
    axs[1,1].imshow(warped_target_image)
    plt.show()




import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
from skimage import io
import numpy as np
from skimage.util import img_as_ubyte
from skimage import color
import numpy as np
from .manipulate import automaticBrightnessAndContrast
import cv2

def showSegmentation(labeled_image: np.array, original_image: np.array = None, save=True, plot=False):
    colored_image = color.label2rgb(labeled_image, bg_label=0)
    if original_image is not None:
        original_image = img_as_ubyte(original_image)

    colored_image_on_DAPI = color.label2rgb(labeled_image, original_image, bg_label=0)
    if save:
        io.imsave("labeled_image.tif", colored_image_on_DAPI)

    if plot:
        plt.imshow(colored_image_on_DAPI)
        plt.axis("off")
        plt.show()

def showImage(image: np.array):
    plt.imshow(image)
    plt.axis("off")
    plt.show()


def createComposite(img1_path, img2_path):
    A = cv2.cv2.imread(img1_path, cv2.cv2.IMREAD_ANYDEPTH)
    B = cv2.cv2.imread(img2_path, cv2.cv2.IMREAD_ANYDEPTH)
    bitDepth = A.dtype
    # print(A.dtype, B.dtype)
    zeros = np.zeros(A.shape[:2], dtype=bitDepth)

    ##Don't forget, cv2.cv2 works with a different order sometimes, so it's not RGB it's BGR
    merged = cv2.cv2.merge((A,B,zeros))
    merged = img_as_ubyte(merged)
    return merged

def evaluateRegistration(ref_image, target_image, registered_target, identifier: str = "", plot=True):
    original_situation = createComposite(ref_image, target_image)
    original_situation = img_as_ubyte(original_situation)
    registered_situation = createComposite(ref_image, registered_target)
    registered_situation = img_as_ubyte(registered_situation)

    if identifier:
        io.imsave(f"overlay_registered_{identifier}.tif", registered_situation)
    if plot:
        fig, axs = plt.subplots(1,2)
        axs[0].imshow(original_situation)
        axs[0].set_title("Before Registration")
        red_patch = mpatches.Patch(color='red', label='MS ref')
        green_patch = mpatches.Patch(color='green', label='ST target')
        axs[0].legend(handles=[red_patch, green_patch])
        axs[1].imshow(registered_situation)
        axs[1].set_title("After Registration")
        green_patch_2 = mpatches.Patch(color='green', label='ST moved')
        axs[1].legend(handles=[red_patch, green_patch_2])
        plt.show()




import os
import SimpleITK as sitk
import pandas as pd
import cv2 as cv
import numpy as np
from skimage import io
from skimage.color import rgb2gray
from skimage.util import invert,img_as_ubyte, img_as_uint
from skimage.transform import rotate
import matplotlib.pyplot as plt
from .manipulate import equalizeImageSize
from .register import performRegistration
from .measure import measureLabeledImage
from .decorators import measureTime


def expandBbox(bbox, size: int):
    min_row, min_col, max_row, max_col = bbox
    min_row -= size
    max_row += size
    min_col -= size
    max_col += size
    bbox = min_row, min_col, max_row, max_col
    return bbox

def cutBboxFromImage(bbox, image: np.ndarray):
        min_row, min_col, max_row, max_col = bbox

        cut_image = image[min_row:max_row, min_col:max_col]
        #io.imsave(filename, cut_image)
        return cut_image

def isolateForeground(input_image: np.array, kernel_size:int = 100, bbox_expansion:int= 1000):
    # create labeled image
    # gray_correct = np.array(255 * (input_image / 255) ** 1.2 , dtype='uint8')
    gray_correct = img_as_ubyte(input_image)
    # Local adaptative threshold

    thresh = cv.adaptiveThreshold(gray_correct, 255, cv.ADAPTIVE_THRESH_GAUSSIAN_C, cv.THRESH_BINARY, 255, 19)
    thresh = cv.bitwise_not(thresh)

    # # Dilatation et erosion
    kernel = np.ones((kernel_size,kernel_size), np.uint8)
    img_dilation = cv.dilate(thresh, kernel, iterations=1)
    img_erode = cv.erode(img_dilation,kernel, iterations=1)
    # # clean all noise after dilatation and erosion
    img_erode = cv.medianBlur(img_erode, 7)

    ret, labeled_image = cv.connectedComponents(img_erode)


    # io.imsave("tmp_labeledimg.tif", labeled_image)
    # Meaure imageprops
    df = measureLabeledImage(labeled_image)

    # get max area row
    max_area_label_number = df.iloc[df['Area'].idxmax()]['image_label']

    # check if the object is horizontal or vertical, horizontal if orientation < 0
    bool_rotate = df.iloc[df['Area'].idxmax()]['orientation']

    # Set everything that isnt max label to zero
    labeled_image[labeled_image!=max_area_label_number] = 0
    image_bbox =  df.iloc[df['Area'].idxmax()]['bbox']
    image_bbox = expandBbox(image_bbox, bbox_expansion)

    foreground_image = cutBboxFromImage(image_bbox, input_image)

    return foreground_image, bool_rotate

def ST_to_MSI_registration(fixed_path: str, moving_path:str):
    """ST_to_MSI_registration.

    BE WARNED. This function is CPU and Memory intensive!

    """

    # fixed_image = sitk.ReadImage(fixed_path, sitk.sitkFloat32)
    # moving_image = sitk.ReadImage(moving_path, sitk.sitkFloat32)

    out_file_base = os.path.splitext(moving_path)[0]

    ref_image = io.imread(fixed_path)
    target_image = io.imread(moving_path)

    ref_image = invert(ref_image)

    target_image = invert(target_image)


    # Crop foreground from the weird Spatial H&E stain
    target_image, bool_rotate = isolateForeground(target_image, bbox_expansion= 500)

    if bool_rotate:
        target_image = rotate(target_image, -90, resize=True)

    # Downscale
    target_image = equalizeImageSize(ref_image, target_image)

    # Equalize ref image dtype, cause target image comes out as uint
    ref_image = img_as_uint(ref_image)

    fixed_image = sitk.GetImageFromArray(ref_image)

    moving_image = sitk.GetImageFromArray(target_image)

    fixed_image = sitk.Cast(fixed_image, sitk.sitkFloat32)
    moving_image = sitk.Cast(moving_image, sitk.sitkFloat32)
    
    initial_transform = sitk.CenteredTransformInitializer(
    fixed_image,
    moving_image,
    sitk.AffineTransform(2),
    sitk.CenteredTransformInitializerFilter.GEOMETRY,
    )

    moving_resampled = sitk.Resample(
        moving_image,
        fixed_image,
        initial_transform,
        sitk.sitkLinear,
        0.0,
        moving_image.GetPixelID(),
    )
    registration_method = sitk.ImageRegistrationMethod()

    # Similarity metric settings.
    registration_method.SetMetricAsMattesMutualInformation(numberOfHistogramBins=50)
    registration_method.SetMetricSamplingStrategy(registration_method.RANDOM)
    registration_method.SetMetricSamplingPercentage(0.10)

    registration_method.SetInterpolator(sitk.sitkLinear)

    # Optimizer settings.
    registration_method.SetOptimizerAsGradientDescent(
        learningRate=1.0,
        numberOfIterations=50000,
        convergenceMinimumValue=1e-6,
        convergenceWindowSize=10,
    )
    registration_method.SetOptimizerScalesFromPhysicalShift()

    # Setup for the multi-resolution framework.
    registration_method.SetShrinkFactorsPerLevel(shrinkFactors=[4, 2, 1])
    registration_method.SetSmoothingSigmasPerLevel(smoothingSigmas=[2, 1, 0])
    registration_method.SmoothingSigmasAreSpecifiedInPhysicalUnitsOn()

    # Don't optimize in-place, we would possibly like to run this cell multiple times.
    registration_method.SetInitialTransform(initial_transform)

    intermediate_transform = registration_method.Execute(
        sitk.Cast(fixed_image, sitk.sitkFloat32), sitk.Cast(moving_image, sitk.sitkFloat32)
    )
    print(f"Affine Final metric value: {registration_method.GetMetricValue()}")
    print(f"Affine Optimizer's stopping condition, {registration_method.GetOptimizerStopConditionDescription()}")


    bspline_registration_method = sitk.ImageRegistrationMethod()

    transformDomainMeshSize = [8] * moving_resampled.GetDimension()
    tx = sitk.BSplineTransformInitializer(fixed_image, transformDomainMeshSize)


    bspline_registration_method.SetMetricAsCorrelation()

    bspline_registration_method.SetOptimizerAsLBFGSB(gradientConvergenceTolerance=1e-5,
                               numberOfIterations=100,
                               maximumNumberOfCorrections=5,
                               maximumNumberOfFunctionEvaluations=1000,
                               costFunctionConvergenceFactor=1e+7)
    bspline_registration_method.SetInitialTransform(tx, inPlace=False)
    bspline_registration_method.SetInterpolator(sitk.sitkLinear)

    #bspline_registration_method.SetOptimizerScalesFromPhysicalShift()

    # Setup for the multi-resolution framework.
    #bspline_registration_method.SetShrinkFactorsPerLevel(shrinkFactors=[4, 2, 1])
    #bspline_registration_method.SetSmoothingSigmasPerLevel(smoothingSigmas=[2, 1, 0])
    #bspline_registration_method.SmoothingSigmasAreSpecifiedInPhysicalUnitsOn()


    final_transform = bspline_registration_method.Execute(
        sitk.Cast(fixed_image, sitk.sitkFloat32), sitk.Cast(moving_resampled, sitk.sitkFloat32)
    )
    print(f"Bspline Final metric value: {bspline_registration_method.GetMetricValue()}")
    print(f"Bspline Optimizer's stopping condition, {bspline_registration_method.GetOptimizerStopConditionDescription()}")

    moving_bspline_resampled = sitk.Resample(
        moving_resampled,
        fixed_image,
        final_transform,
        sitk.sitkLinear,
        0.0,
        moving_image.GetPixelID(),
    )
    sitk.WriteImage(
        moving_bspline_resampled, os.path.join(f"{out_file_base}_registered.tif")
    )
    sitk.WriteTransform(
        final_transform, os.path.join(f"{out_file_base}_transform.txt")
    )

    return moving_bspline_resampled, final_transform





import numpy as np

def getPatch(arr, idx, radius=3, fill=None):
    """
    Gets surrounding elements from a numpy array

    Parameters:
    arr (ndarray of rank N): Input array
    idx (N-Dimensional Index): The index at which to get surrounding elements. If None is specified for a particular axis,
        the entire axis is returned.
    radius (array-like of rank N or scalar): The radius across each axis. If None is specified for a particular axis,
        the entire axis is returned.
    fill (scalar or None): The value to fill the array for indices that are out-of-bounds.
        If value is None, only the surrounding indices that are within the original array are returned.

    Returns:
    ndarray: The surrounding elements at the specified index
    """

    assert len(idx) == len(arr.shape)

    if np.isscalar(radius):
        radius = tuple([radius for i in range(len(arr.shape))])

    slices = []
    paddings = []
    for axis in range(len(arr.shape)):
        if idx[axis] is None or radius[axis] is None:
            slices.append(slice(0, arr.shape[axis]))
            paddings.append((0, 0))
            continue

        r = radius[axis]
        l = idx[axis] - r
        r = idx[axis] + r

        pl = 0 if l > 0 else abs(l)
        pr = 0 if r < arr.shape[axis] else r - arr.shape[axis] + 1

        slices.append(slice(max(0, l), min(arr.shape[axis], r+1)))
        paddings.append((pl, pr))

    if fill is None:
        return arr[tuple(slices)]
    return np.pad(arr[tuple(slices)], paddings, 'constant', constant_values=fill)

def getLastDimensions(array, nr_dims: int = 2):
    """Extracts last X dimensions from an array

    Parameters
    ----------
    array :
        input array to be extracted
    nr_dims : int
        nr of dimensions to extract. Default = 2 
    """
    if array.ndim <= nr_dims:
        return array
    if nr_dims > nr_dims:
        raise ValueError(f"Requested nr of dimensions {nr_dims} higher than nr of dimensions present in array.")
    slc = [0] * (array.ndim - nr_dims)
    slc += [slice(None) for i in range(0, nr_dims)]
    return array[tuple(slc)]





from skimage.io import imread_collection, imsave, imread
import numpy as np
from tifffile import imsave

from typing import List, Tuple
import glob
import os

def calculateOptimalLargestResolution(glob_pattern: str, target_tile_height: int, target_tile_width: int) -> List[int]: 
    """Calculates the optimal maximum resolution based on a set of images assuming that it will be used to tile the images into equal tiles of a certain size, given by the input variables.


    Parameters
    ----------
    glob_pattern : str
        glob pattern hitting all images to be taken into account for the calculation.
    target_tile_height : int
        target_tile_height
    target_tile_width : int
        target_tile_width

    Returns
    -------
    List[int]

    """
    images_array = np.array(imread_collection(glob_pattern), dtype = object)

    heights = []
    widths = []
    for i in range(len(images_array)):
        heights.append(images_array[i].shape[0])
        widths.append(images_array[i].shape[1])
    max_rows = max(heights) 
    max_columns = max(widths)

    ydiv = np.ceil(max_rows / target_tile_height)
    xdiv = np.ceil(max_columns / target_tile_width)

    target_full_rows = ydiv * target_tile_height
    target_full_columns = xdiv * target_tile_width

    return target_full_rows, target_full_columns, ydiv, xdiv

def padImage(image: np.array, target_full_rows: int, target_full_columns: int) -> np.ndarray:
    """Pads an image with black pixels up until a certain number of rows and columns

    Parameters
    ----------
    image : np.array
        image to be padded
    target_full_rows : int
        Number of rows to be achieved by padding
    target_full_columns : int
        Number of columns to be achieved by padding

    Returns
    -------
    np.ndarray

    """
    rowdiff = target_full_rows - image.shape[0]
    columndiff = target_full_columns - image.shape[1]
    padded_img = np.pad(image, ((0, rowdiff), (0, columndiff)))
    return padded_img

def tileImage(image: np.ndarray, ydiv: int, xdiv: int, image_prefix: str="image_tile_"):
    """Tile an image into smaller tiles based on divisions in x and y axes, and saving them to tif files.

    Parameters
    ----------
    image : np.ndarray
        image
    ydiv : int
        Nr of divisions to make in the y axis
    xdiv : int
        Nr of divisions to make in the x axis
    image_prefix : str
        prefix to add to the new filenames of the tiles. default = image_tile
    """
    if not image_prefix.endswith("_"):
        image_prefix += "_"

    temp_split = np.array_split(image, ydiv, axis = 0)
    # Item sublist part is just to unpack a list of lists into one list
    final_split = [item for sublist in [np.array_split(row, xdiv, axis = 1) for row in temp_split] for item in sublist]

    for i, img in enumerate(final_split, 1):
        imsave(f"{image_prefix}tile{i}.tif", img)

def tile(glob_pattern: str, target_tile_width: int, target_tile_height: int, out_dir: str = "") -> Tuple[int]:
    """Tile the images caught by the glob pattern by first padding them to a global image size needed to tile them all into the same tile size, given by the input values. Tiles are written to tif files in the basedir of the original image, or to out_dir if given, with naming convention {out_dir | basedir}/{image_name}_tile{i}.tif

    Parameters
    ----------
    glob_pattern : str
        glob_pattern that catches all images to be tiled.
    target_tile_width : int
        Width of the tiles
    target_tile_height : int
        Height of the tiles

    Returns
    -------
    Tuple[int]

    """
    if out_dir:
        os.makedirs(out_dir, exist_ok=True)

    target_full_rows, target_full_columns, ydiv, xdiv = calculateOptimalLargestResolution(glob_pattern, target_tile_height, target_tile_width)

    padded_imgs = {}
    if not out_dir:
        for image_path in glob.glob(glob_pattern):
            padded_imgs[os.path.splitext(image_path)[0]] = padImage(imread(image_path), int(target_full_rows), int(target_full_columns))
    else:
        for image_path in glob.glob(glob_pattern):
            padded_imgs[os.path.join(out_dir, os.path.splitext(os.path.basename(image_path))[0])] = padImage(imread(image_path), int(target_full_rows), int(target_full_columns))

    for k, padded_img in padded_imgs.items():
        tileImage(padded_img,ydiv = ydiv, xdiv = xdiv, image_prefix = k)
    return xdiv, ydiv

if __name__ == '__main__':
    tile("/home/david/.config/nnn/mounts/nacho@10.38.76.144/tool/enteric_neurones/slidescanner_examples/Good/Slide2-2-2_Region0000_Channel647,555,488_Seq0017/Slide2-2-2_Region0000_Channel647,555,488_Seq0017_555.tif", 2000, 2000, out_dir = "/home/david/.config/nnn/mounts/nacho@10.38.76.144/tool/enteric_neurones/slidescanner_examples/Good/Slide2-2-2_Region0000_Channel647,555,488_Seq0017/Slide2-2-2_Region0000_Channel647,555,488_Seq0017_555_tiles/")
